from flask import Response


class RequestFilter:
    def before_request(self):
        pass

    def after_request(self, response: Response):
        return response


class RequestFilterChain(RequestFilter):
    def __init__(self):
        self._filters = []

    def add_request_filter(self, request_filter: RequestFilter):
        self._filters.append(request_filter)

    def before_request(self):
        for f in self._filters:
            if f.before_request:
                f.before_request()

    def after_request(self, response):
        for f in self._filters:
            if f.after_request:
                response = f.after_request(response)
        return response



import base64
import io
import json
import logging
from typing import List, Union, Optional
from urllib.parse import urlencode, urlsplit, urlunsplit

import dns.message
import dns.query
import dns.rdatatype
import requests
import yaml

from guniflask.config.app_settings import Settings, settings
from guniflask.context.annotation import condition_on_setting, configuration
from guniflask.service_discovery.config import ServiceDiscoveryConfigurer
from guniflask.service_discovery.discovery_client import DiscoveryClient
from guniflask.service_discovery.errors import ServiceDiscoveryError
from guniflask.service_discovery.load_balancer_client import LoadBalancerClient
from guniflask.service_discovery.service_instance import ServiceInstance

log = logging.getLogger(__name__)


class ConsulClientError(ServiceDiscoveryError):
    pass


class ConsulClient(DiscoveryClient, LoadBalancerClient):
    api_version = 'v1'

    def __init__(
            self,
            host: str = '127.0.0.1',
            port: int = 8500,
            dns_port: int = 8600,
            scheme: str = 'http',
            config_key: str = None,
            config_format: str = 'yaml',
    ):
        self.host = host
        self.port = port
        self.dns_port = dns_port
        self.scheme = scheme
        self.session = requests.Session()
        self.base_url = f'{scheme}://{host}:{port}/{self.api_version}'
        self.config_key = config_key
        self.config_format = config_format.lower()

        assert self.config_format in {'yaml', 'json'}, \
            f'unsupported Consul configuration format: {self.config_format}'

    def register_service(self, name: str,
                         service_id: str = None,
                         tags: List[str] = None,
                         address: str = None,
                         port: int = None,
                         check: dict = None):
        api_path = '/agent/service/register'
        args = {}
        if check is not None:
            args['replace-existing-checks'] = 'true'
        if len(args) > 0:
            api_path = f'{api_path}?{urlencode(args)}'

        data = {
            'Name': name,
            'ID': service_id,
            'Tags': tags,
            'Address': address,
            'Port': port,
            'Check': check
        }
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.put(url, json=data)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)

    def deregister_service(self, service_id: str):
        api_path = f'/agent/service/deregister/{service_id}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.put(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)

    def get_service_by_id(self, service_id: str):
        api_path = f'/agent/service/{service_id}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)
        return resp.json()

    @staticmethod
    def http_check(
            name: str,
            url: str,
            check_id: str = None,
            interval: str = None,
            deregister_after: str = None):
        return {
            'Name': name,
            'CheckID': check_id,
            'HTTP': url,
            'Interval': interval,
            'DeregisterCriticalServiceAfter': deregister_after,
        }

    def get_service_instances(self, service_name: str) -> List[ServiceInstance]:
        api_path = f'/agent/health/service/name/{service_name}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)
        data = resp.json()
        services = []
        for d in data:
            if d['AggregatedStatus'] == 'passing':
                s = d['Service']
                services.append(ServiceInstance(service_id=s['ID'],
                                                host=s['Address'],
                                                port=s['Port']))
        return services

    def choose(self, service_name: str) -> Union[ServiceInstance, None]:
        request = dns.message.make_query(f'{service_name}.service.consul', dns.rdatatype.SRV)
        response = dns.query.udp(request, self.host, port=self.dns_port)
        if len(response.answer) > 0:
            answer = response.answer[0]
            port = None
            target = None
            for k in answer:
                port = k.port
                target = k.target
                break
            if target is not None:
                for additional in response.additional:
                    if additional.name == target:
                        for k in additional:
                            return ServiceInstance(host=k.address, port=port)

    def reconstruct_url(self, service_instance: ServiceInstance, original_url: str) -> str:
        result = urlsplit(original_url)
        result = result._replace(netloc=f'{service_instance.host}:{service_instance.port}')
        return urlunsplit(result)

    def get_configuration(self) -> Optional[dict]:
        if not self.config_key:
            return

        api_path = f'/kv/{self.config_key}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)
        data = resp.json()[0]['Value']
        s = base64.b64decode(data)
        if self.config_format == 'yaml':
            return yaml.safe_load(io.BytesIO(s))
        if self.config_format == 'json':
            return json.load(io.BytesIO(s))


@condition_on_setting('guniflask.consul')
@configuration
class ConsulConfigurer(ServiceDiscoveryConfigurer):

    def __init__(self):
        self._discovery_client: Optional[DiscoveryClient] = None
        self._load_balancer_client: Optional[LoadBalancerClient] = None
        self._service_name: Optional[str] = None
        self._consul_client: Optional[ConsulClient] = None

    @property
    def discovery_client(self) -> DiscoveryClient:
        return self._discovery_client

    @property
    def load_balancer_client(self) -> LoadBalancerClient:
        return self._load_balancer_client

    def configure(self, service_name: str, app_settings: Settings):
        config = app_settings.get_by_prefix('guniflask.consul')
        self._service_name = service_name
        self._register(config, app_settings)

        remote_settings = self._consul_client.get_configuration()
        if remote_settings:
            settings.merge(remote_settings)

    def _register(self, config: dict, app_settings: Settings):
        self._consul_client = ConsulClient(**config)
        self._discovery_client = self._consul_client
        self._load_balancer_client = self._consul_client
        self._do_register(self._consul_client, app_settings)

    def _do_register(self, consul: ConsulClient, app_settings: Settings):
        local_ip = app_settings['ip_address']
        port = app_settings['port']
        service_id = f'{app_settings["app_name"]}-{local_ip}-{port}'
        heath_url = f'http://{local_ip}:{port}/health?' \
                    f'app_name={app_settings["app_name"]}'
        try:
            consul.register_service(
                app_settings['app_name'],
                service_id=service_id,
                address=local_ip,
                port=port,
                check=consul.http_check(
                    service_id,
                    heath_url,
                    check_id=service_id,
                    interval='10s',
                    deregister_after='10m',
                )
            )
        except ConsulClientError as e:
            log.error('Failed to register service to Consul at %s:%s: %s', consul.host, consul.port, e)



import inspect
from typing import Type, Collection, Union, Any

from guniflask.annotation import Annotation, AnnotationUtils, AnnotationMetadata
from guniflask.config.app_settings import settings
from guniflask.context.condition import Condition, ConditionContext


class Bean(Annotation):
    def __init__(self, name=None):
        super().__init__(name=name)


def bean(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Bean(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Autowired(Annotation):
    def __init__(self):
        super().__init__()


def autowired(func):
    AnnotationUtils.add_annotation(func, Autowired())
    return func


class Component(Annotation):
    def __init__(self, name=None):
        super().__init__(name=name)


def component(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Component(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Configuration(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def configuration(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Configuration(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Conditional(Annotation):
    def __init__(self, condition: Type[Condition]):
        super().__init__(condition=condition)


def conditional(condition: Union[Type[Condition], Condition]):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Conditional(condition))
        return func

    return wrap_func


class SettingCondition(Condition):
    def __init__(self, name: str, value: Any = None):
        self.name = name
        self.value = value

    def matches(self, context: ConditionContext, metadata: AnnotationMetadata) -> bool:
        v = settings.get_by_prefix(self.name)
        if self.value is None:
            return v is not None
        return v == self.value


def condition_on_setting(name: str, value: Any = None):
    return conditional(SettingCondition(name, value=value))


class Repository(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def repository(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Repository(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Service(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def service(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Service(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Controller(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def controller(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Controller(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Include(Annotation):
    def __init__(self, values: Collection = None):
        super().__init__(values=values)


def include(*values):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Include(values))
        return func

    return wrap_func



import pytest
from sqlalchemy import create_engine, Column, String, Integer, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship, sessionmaker

from guniflask.orm import result_to_dict, BaseModelMixin

Base = declarative_base()


class User(BaseModelMixin, Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    nickname = Column(String)


class Article(BaseModelMixin, Base):
    __tablename__ = 'article'
    id = Column(Integer, primary_key=True)
    title = Column(String)
    content = Column(String)
    user_id = Column(Integer, ForeignKey('user.id'))
    author = relationship('User', backref=backref('articles'))


@pytest.fixture
def session():
    engine = create_engine('sqlite:///:memory:')
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    session = Session()

    user = User(id=1, name='Bob', nickname='Good Boy')
    session.add(user)
    article = Article(id=1, title='Title', content='Content', author=user)
    session.add(article)
    session.commit()

    return session


def test_result_to_dict(session):
    result = session.query(User).with_entities(User.name, User.nickname).filter_by(name='Bob').all()
    assert result_to_dict(result[0]) == {'name': 'Bob', 'nickname': 'Good Boy'}


def test_model_from_dict():
    article = Article.from_dict(dict(id=1, title='Title'))
    assert article.id == 1
    assert article.title == 'Title'


def test_model_from_dict_with_many_to_one_relation():
    article_data = dict(
        id=1,
        title='Title',
        content='Content',
        user_id=1,
        author=dict(
            id=1,
            name='Bob',
            nickname='Good Boy',
        ),
    )
    article = Article.from_dict(article_data)
    assert article.id == 1
    assert article.title == 'Title'
    assert article.content == 'Content'
    assert article.user_id == 1
    assert article.author.id == 1
    assert article.author.name == 'Bob'
    assert article.author.nickname == 'Good Boy'


def test_model_from_dict_with_one_to_many_relation():
    user_data = dict(
        id=1,
        name='Bob',
        nickname='Good Boy',
        articles=[
            dict(
                id=1,
                title='Title',
                content='Content',
                user_id=1,
            )
        ],
    )
    user = User.from_dict(user_data)
    assert user.id == 1
    assert user.name == 'Bob'
    assert user.nickname == 'Good Boy'
    article = user.articles[0]
    assert article.id == 1
    assert article.title == 'Title'
    assert article.content == 'Content'
    assert article.user_id == 1


def test_model_from_dict_with_ignore():
    article = Article.from_dict(dict(id=1, title='Title'), ignore='id')
    assert article.id is None
    assert article.title == 'Title'


def test_model_from_dict_with_only():
    article = Article.from_dict(dict(id=1, title='Title', content='Content'), only='title,content')
    assert article.id is None
    assert article.title == 'Title'
    assert article.content == 'Content'


def test_model_to_dict():
    article = Article(id=1, title='Title', content='Content', user_id=1,
                      author=User(id=1, name='Bob', nickname='Good Boy'))
    d = article.to_dict(include='author')
    assert d == {'id': 1, 'title': 'Title', 'content': 'Content', 'user_id': 1,
                 'author': {'id': 1, 'name': 'Bob', 'nickname': 'Good Boy'}}


def test_model_to_dict_with_ignore():
    article = Article(id=1, title='Title', content='Content', user_id=1,
                      author=User(id=1, name='Bob', nickname='Good Boy'))
    d = article.to_dict(ignore='id,user_id,author.id', include='author')
    assert d == {'title': 'Title', 'content': 'Content',
                 'author': {'name': 'Bob', 'nickname': 'Good Boy'}}


def test_model_to_dict_with_only():
    article = Article(id=1, title='Title', content='Content', user_id=1,
                      author=User(id=1, name='Bob', nickname='Good Boy'))
    d = article.to_dict(only='title,content,author.name,author.nickname', include='author')
    assert d == {'title': 'Title', 'content': 'Content',
                 'author': {'name': 'Bob', 'nickname': 'Good Boy'}}


def test_model_to_dict_with_one_to_many_relation(session):
    user = session.query(User).filter_by(id=1).first()
    d = user.to_dict(include='articles')
    assert d == {'id': 1, 'name': 'Bob', 'nickname': 'Good Boy',
                 'articles': [{'id': 1, 'title': 'Title', 'content': 'Content', 'user_id': 1}]}


def test_model_to_dict_with_many_to_one_relation(session):
    article = session.query(Article).filter_by(id=1).first()
    d = article.to_dict(include='author')
    assert d == {'id': 1, 'title': 'Title', 'content': 'Content', 'user_id': 1,
                 'author': {'id': 1, 'name': 'Bob', 'nickname': 'Good Boy'}}


def test_update_model_by_dict():
    article = Article()
    article.update_by_dict(dict(id=1, title='Title', author=dict(id=1, name='Bob')))
    assert article.id == 1
    assert article.title == 'Title'
    assert article.author.id == 1
    assert article.author.name == 'Bob'


def test_update_model_by_dict_with_ignore():
    article = Article(author=User())
    article.update_by_dict(dict(id=1, title='Title', author=dict(id=1, name='Bob')), ignore='id,author.id')
    assert article.id is None
    assert article.title == 'Title'
    assert article.author.id is None
    assert article.author.name == 'Bob'


def test_update_model_by_dict_with_only():
    article = Article()
    article.update_by_dict(
        dict(
            id=1,
            title='Title',
            content='Content',
            author=dict(id=1, name='Bob'),
        ),
        only='title,content,author.name',
    )
    assert article.id is None
    assert article.title == 'Title'
    assert article.content == 'Content'
    assert article.author.id is None
    assert article.author.name == 'Bob'


def test_update_model_by_dict_with_list():
    user = User()
    with pytest.raises(RuntimeError):
        user.update_by_dict({'name': 'Bob', 'articles': [{'title': 'Title'}]})



from typing import Union, List, Set

from sqlalchemy.orm import Query

from guniflask.orm.model_utils import dict_to_model, model_to_dict, update_model_by_dict


class BaseModelMixin:
    query: Query

    @classmethod
    def from_dict(
            cls,
            dict_obj: dict,
            ignore: Union[str, List[str], Set[str]] = None,
            only: Union[str, List[str], Set[str]] = None,
    ):
        return dict_to_model(dict_obj, cls, ignore=ignore, only=only)

    def to_dict(
            self,
            ignore: Union[str, List[str], Set[str]] = None,
            only: Union[str, List[str], Set[str]] = None,
            include: Union[str, List[str], Set[str]] = None,
    ):
        return model_to_dict(self, ignore=ignore, only=only, include=include)

    def update_by_dict(
            self,
            dict_obj: dict,
            ignore: Union[str, List[str], Set[str]] = None,
            only: Union[str, List[str], Set[str]] = None,
    ):
        return update_model_by_dict(self, dict_obj, ignore=ignore, only=only)



import fcntl
from os.path import join

from guniflask.config.app_settings import settings
from guniflask.utils.path import make_temp_dir


class ServiceLock:
    locks = {}

    def __init__(self, name: str):
        self.name = name

    def acquire(self) -> bool:
        instance_id = self._generate_instance_id()
        if instance_id in self.locks:
            return True

        temp_dir = make_temp_dir(self.__class__.__name__)
        fd = open(join(temp_dir, instance_id), 'w')
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            return False
        self.locks[instance_id] = fd
        return True

    def _generate_instance_id(self) -> str:
        return f'{settings["app_name"]}.{settings["port"]}.lock'

    def release(self):
        instance_id = self._generate_instance_id()
        if instance_id in self.locks:
            fd = self.locks.pop(instance_id)
            try:
                fd.close()
            except IOError:
                pass



from typing import List

from guniflask.data_model import DataModel


class Person(DataModel):
    name: str
    age: int = None


class Student(Person):
    graduate: bool = None


def test_create_data_model():
    student = Student(name='Bob', graduate=True)
    assert student.name == 'Bob'
    assert student.age is None
    assert student.graduate is True


def test_data_model_from_dict():
    student = Student.from_dict(dict(name='Bob', graduate=True))
    assert student.name == 'Bob'
    assert student.age is None
    assert student.graduate is True


def test_data_model_to_dict():
    d = Student.from_dict(dict(name='Bob', graduate=True)).to_dict()
    assert d['name'] == 'Bob'
    assert d['age'] is None
    assert d['graduate'] is True


class User(DataModel):
    name: str


class Article(DataModel):
    title: str
    author: User = None


def test_data_model_from_dict_with_many_to_one():
    d = Article.from_dict(dict(title='Title', author=dict(name='Bob')))
    assert d.title == 'Title'
    assert d.author.name == 'Bob'


def test_data_model_to_dict_with_many_to_one():
    d = Article.from_dict(dict(title='Title', author=dict(name='Bob'))).to_dict()
    assert d == {
        'title': 'Title',
        'author': {
            'name': 'Bob',
        },
    }


class ArticleSet(DataModel):
    name: str
    articles: List[Article]


def test_data_model_from_dict_with_one_to_many():
    d = ArticleSet.from_dict(dict(name='Name', articles=[dict(title='Title')]))
    assert d.name == 'Name'
    assert len(d.articles) == 1
    assert d.articles[0].title == 'Title'


def test_data_model_to_dict_with_one_to_many():
    article_set = ArticleSet.from_dict(dict(name='Name', articles=[dict(title='Title')]))

    d = article_set.to_dict()
    assert d == {
        'name': 'Name',
        'articles': [
            {
                'title': 'Title',
                'author': None,
            }
        ],
    }

    d = article_set.to_dict(ignore='articles.author')
    assert d == {
        'name': 'Name',
        'articles': [
            {
                'title': 'Title',
            }
        ],
    }

    d = article_set.to_dict(ignore='articles')
    assert d == {
        'name': 'Name',
    }

    d = article_set.to_dict(only='articles.title')
    assert d == {
        'articles': [
            {
                'title': 'Title',
            }
        ]
    }



from guniflask.annotation import AnnotationMetadata
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.context.annotation import Conditional
from guniflask.context.condition import ConditionContext, Condition


class ConditionEvaluator:
    def __init__(self, registry: BeanDefinitionRegistry):
        self.context = ConditionContext(registry)

    def should_skip(self, metadata: AnnotationMetadata):
        if metadata is None or not metadata.is_annotated(Conditional):
            return False
        condition_cls = self.get_condition_class(metadata)
        if isinstance(condition_cls, Condition):
            condition = condition_cls
        else:
            condition = condition_cls()
        if not condition.matches(self.context, metadata):
            return True
        return False

    def get_condition_class(self, metadata: AnnotationMetadata):
        annotation = metadata.get_annotation(Conditional)
        return annotation['condition']



from flask import Flask

from guniflask.beans.definition import BeanDefinition
from guniflask.context.default_bean_context import AnnotationConfigBeanContext
from guniflask.web.blueprint_post_processor import BlueprintPostProcessor
from guniflask.web.config_constants import *


class WebApplicationContext(AnnotationConfigBeanContext):
    def __init__(self, app: Flask):
        super().__init__()
        self.app = app

    def _post_process_bean_factory(self):
        super()._post_process_bean_factory()

        if not self.contains_bean_definition(BLUEPRINT_POST_PROCESSOR):
            bean_definition = BeanDefinition(BlueprintPostProcessor)
            self.register_bean_definition(BLUEPRINT_POST_PROCESSOR, bean_definition)



from importlib import import_module

from flask import Blueprint, Flask

from guniflask.config.app_settings import Settings
from guniflask.config.env import app_name_from_env
from guniflask.config.load_utils import load_app_settings
from guniflask.security_config.authentication_config import AuthenticationConfiguration
from guniflask.security_config.web_security_config import WebSecurityConfiguration
from guniflask.service_discovery.config import ServiceDiscoveryConfiguration
from guniflask.utils.path import walk_modules
from guniflask.web.context import WebApplicationContext
from guniflask.web.scheduling_config import WebAsyncConfiguration, WebSchedulingConfiguration


class AppInitializer:
    def __init__(self):
        self.name = app_name_from_env()
        app_settings = load_app_settings(self.name)
        self.settings = Settings(app_settings)

    def init(self, app=None, with_context=True):
        if app is None:
            app = Flask(self.name)
        self._make_settings(app)
        if with_context:
            self._create_bean_context(app)
        self._init_app(app)
        with app.app_context():
            self._register_blueprints(app)
            self._refresh_bean_context(app)
        return app

    def _make_settings(self, app):
        app_module = self._get_app_module()
        _make_settings = getattr(app_module, 'make_settings', None)
        if _make_settings:
            s = self.settings
            _make_settings(app, s)

    def _init_app(self, app):
        setattr(app, 'settings', self.settings)
        for k, v in self.settings.items():
            if k.isupper():
                app.config[k] = v

        app_module = self._get_app_module()
        _init_app = getattr(app_module, 'init_app', None)
        if _init_app:
            s = self.settings
            _init_app(app, s)

    def _create_bean_context(self, app):
        bean_context = WebApplicationContext(app)
        self._auto_configure_bean_context(bean_context)
        setattr(app, 'bean_context', bean_context)

    def _auto_configure_bean_context(self, bean_context: WebApplicationContext):
        bean_context.register(AuthenticationConfiguration)
        bean_context.register(WebSecurityConfiguration)
        bean_context.register(WebAsyncConfiguration)
        bean_context.register(WebSchedulingConfiguration)
        bean_context.register(ServiceDiscoveryConfiguration)

    def _refresh_bean_context(self, app):
        if hasattr(app, 'bean_context'):
            bean_context: WebApplicationContext = getattr(app, 'bean_context')
            bean_context.scan(self.settings['app_name'])
            bean_context.refresh()

    def _register_blueprints(self, app):
        """
        Register declared Blueprint
        """
        registered_blueprints = set()

        def iter_blueprints():
            for module in walk_modules(self.settings['app_name']):
                for obj in vars(module).values():
                    if isinstance(obj, Blueprint) and obj not in registered_blueprints:
                        yield obj
                        registered_blueprints.add(obj)

        for b in iter_blueprints():
            app.register_blueprint(b)

        del registered_blueprints

    def _get_app_module(self):
        return import_module(self.settings['app_name'] + '.app')



import logging
from abc import ABCMeta, abstractmethod

from guniflask.config.app_settings import settings, Settings
from guniflask.context.annotation import configuration, bean, include
from guniflask.distributed.local_lock import ServiceLock
from guniflask.service_discovery.discovery_client import DiscoveryClient
from guniflask.service_discovery.health_endpoint import HealthEndpoint
from guniflask.service_discovery.load_balancer_client import LoadBalancerClient, RestClient

log = logging.getLogger(__name__)


class ServiceDiscoveryConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def configure(self, service_name: str, app_settings: Settings) -> bool:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def discovery_client(self) -> DiscoveryClient:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def load_balancer_client(self) -> LoadBalancerClient:
        pass  # pragma: no cover


@configuration
@include('guniflask.service_discovery.consul.ConsulConfigurer')
class ServiceDiscoveryConfiguration:

    def __init__(self, service_discovery_configurer: ServiceDiscoveryConfigurer = None):
        self._service_discovery_configurer = service_discovery_configurer

        self._register_lock = ServiceLock('__service_registration_lock')
        self._auto_register()

    @bean
    def heath_endpoint(self) -> HealthEndpoint:
        endpoint = HealthEndpoint()
        return endpoint

    @bean
    def discovery_client(self) -> DiscoveryClient:
        if self._service_discovery_configurer:
            return self._service_discovery_configurer.discovery_client

    @bean
    def load_balancer_client(self) -> LoadBalancerClient:
        if self._service_discovery_configurer:
            return self._service_discovery_configurer.load_balancer_client

    @bean
    def load_balanced_request(self) -> RestClient:
        if self._service_discovery_configurer:
            load_balancer_client = self._service_discovery_configurer.load_balancer_client
            if load_balancer_client is not None:
                return RestClient(load_balancer_client)

    def _auto_register(self):
        if not self._register_lock.acquire():
            return

        app_settings = settings._get_current_object()
        service_name = app_settings['app_name']
        if self._service_discovery_configurer is not None:
            self._service_discovery_configurer.configure(service_name, app_settings)



from abc import ABCMeta, abstractmethod
from typing import List

from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.beans.singleton_registry import SingletonBeanRegistry


class BeanFactory(metaclass=ABCMeta):
    @abstractmethod
    def get_bean(self, bean_name: str, required_type: type = None):
        pass  # pragma: no cover

    @abstractmethod
    def get_bean_of_type(self, required_type: type):
        pass  # pragma: no cover

    @abstractmethod
    def get_beans_of_type(self, required_type: type):
        pass  # pragma: no cover

    @abstractmethod
    def is_type_match(self, bean_name: str, type_to_match: type) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def get_bean_names_for_type(self, required_type: type) -> List[str]:
        pass  # pragma: no cover


class BeanNameAware(metaclass=ABCMeta):
    @abstractmethod
    def set_bean_name(self, bean_name: str):
        pass  # pragma: no cover


class BeanFactoryAware(metaclass=ABCMeta):
    @abstractmethod
    def set_bean_factory(self, bean_factory: BeanFactory):
        pass  # pragma: no cover


class ConfigurableBeanFactory(SingletonBeanRegistry, BeanFactory, metaclass=ABCMeta):

    @abstractmethod
    def add_bean_post_processor(self, bean_post_processor: BeanPostProcessor):
        pass  # pragma: no cover

    @abstractmethod
    def pre_instantiate_singletons(self):
        pass  # pragma: no cover



from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.basic_authentication_filter import BasicAuthenticationFilter
from guniflask.security_config.http_security_builder import HttpSecurityBuilder
from guniflask.security_config.security_configurer import SecurityConfigurer


class HttpBasicConfigurer(SecurityConfigurer):

    def configure(self, http: HttpSecurityBuilder):
        authentication_manager = http.get_shared_object(AuthenticationManager)
        basic_auth_filter = BasicAuthenticationFilter(authentication_manager)
        http.add_request_filter(basic_auth_filter)



from abc import ABCMeta, abstractmethod
from typing import List

from guniflask.beans.definition import BeanDefinition


class BeanDefinitionRegistry(metaclass=ABCMeta):

    @abstractmethod
    def register_bean_definition(self, bean_name: str, bean_definition: BeanDefinition):
        pass  # pragma: no cover

    @abstractmethod
    def get_bean_definition(self, bean_name: str) -> BeanDefinition:
        pass  # pragma: no cover

    @abstractmethod
    def get_bean_definition_names(self) -> List[str]:
        pass  # pragma: no cover

    @abstractmethod
    def contains_bean_definition(self, bean_name: str) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def remove_bean_definition(self, bean_name: str):
        pass  # pragma: no cover



import datetime as dt

from guniflask.annotation import AnnotationUtils
from guniflask.beans.factory import BeanFactory, BeanFactoryAware
from guniflask.beans.lifecycle import SmartInitializingSingleton
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.scheduling.annotation import Scheduled
from guniflask.scheduling.task_scheduler import TaskScheduler


class ScheduledPostProcessor(BeanPostProcessor, BeanFactoryAware, SmartInitializingSingleton):

    def __init__(self, task_scheduler: TaskScheduler = None):
        self.bean_factory = None
        self.task_scheduler = task_scheduler
        self._scheduled_tasks = []

    def set_bean_factory(self, bean_factory: BeanFactory):
        self.bean_factory = bean_factory

    def post_process_after_initialization(self, bean, bean_name: str):
        for m in dir(bean):
            method = getattr(bean, m)
            a = AnnotationUtils.get_annotation(method, Scheduled)
            if a is not None:
                self._scheduled_tasks.append((a, method))
        return bean

    def after_singletons_instantiated(self):
        self._finish_registration()

    def _finish_registration(self):
        if self.task_scheduler is None:
            assert self.bean_factory is not None and isinstance(self.bean_factory, BeanFactory), \
                'Bean factory must be set to find task scheduler'
            self.task_scheduler = self.bean_factory.get_bean_of_type(TaskScheduler)

        for scheduled, method in self._scheduled_tasks:
            method = self._post_process_scheduled_method(method)
            self._schedule_task(scheduled, method)

    def _schedule_task(self, scheduled: Scheduled, method):
        start_time = None
        if scheduled['initial_delay'] is not None:
            start_time = dt.datetime.now() + dt.timedelta(seconds=scheduled['initial_delay'])
        if scheduled['cron'] is not None:
            self.task_scheduler.schedule_with_cron(method, scheduled['cron'], start_time=start_time)
        elif scheduled['interval'] is not None:
            self.task_scheduler.schedule_with_fixed_interval(method, scheduled['interval'], start_time=start_time)
        else:
            self.task_scheduler.schedule(method, start_time=start_time)

    def _post_process_scheduled_method(self, method):
        return method



from typing import Union, List, Set


def _make_rule_set(rule: Union[str, List[str], Set[str]]):
    if isinstance(rule, str):
        rule = [i.strip() for i in rule.split(',')]
    return set(rule or [])


def make_ignore_rule_for_field(field: Union[str, List[str], Set[str]]):
    if isinstance(field, set):
        return field
    return _make_rule_set(field)


def make_include_rule_for_field(field: Union[str, List[str], Set[str]]) -> set:
    if isinstance(field, set):
        return field
    return _make_rule_set(field)


def make_only_rule_for_field(field: Union[str, List[str], Set[str]]) -> set:
    if isinstance(field, set):
        return field
    s = _make_rule_set(field)
    for k in list(s):
        if '.' in k:
            t = k.split('.')
            for i in range(1, len(t)):
                s.add('.'.join(t[:i]))
    return s



import logging
from typing import List

from guniflask.beans.default_factory import DefaultBeanFactory
from guniflask.beans.factory_post_processor import BeanFactoryPostProcessor, BeanDefinitionRegistryPostProcessor
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.context.annotation_config_registry import AnnotatedBeanDefinitionReader, ModuleBeanDefinitionScanner
from guniflask.context.annotation_config_registry import AnnotationConfigRegistry
from guniflask.context.bean_context import BeanContext
from guniflask.context.config_constants import *
from guniflask.context.context_aware_processor import BeanContextAwareProcessor
from guniflask.context.event import ApplicationEvent, ContextRefreshedEvent, ContextClosedEvent
from guniflask.context.event_listener import ApplicationEventListener
from guniflask.context.event_publisher import ApplicationEventPublisher

log = logging.getLogger(__name__)


class DefaultBeanContext(DefaultBeanFactory, BeanContext):
    def __init__(self):
        DefaultBeanFactory.__init__(self)
        self._bean_factory_post_processors = []
        self._app_listeners = set()
        self._app_event_publisher = None

    def add_bean_factory_post_processor(self, post_processor: BeanFactoryPostProcessor):
        self._bean_factory_post_processors.append(post_processor)

    @property
    def bean_factory_post_processors(self) -> List[BeanFactoryPostProcessor]:
        return self._bean_factory_post_processors

    def refresh(self):
        self._prepare_bean_factory()
        try:
            self._post_process_bean_factory()
            self._invoke_bean_factory_post_processors()
            self._register_bean_post_processors()
            self._init_application_event_publisher()
            self._register_application_listeners()
            self._finish_bean_factory_initialization()
            self._finish_refresh()
        except Exception:
            log.error('Exception encountered during context initialization', exc_info=True)
            self._destroy_beans()
            raise

    def close(self):
        self.publish_event(ContextClosedEvent(self))
        self._destroy_beans()

    def add_application_listener(self, listener: ApplicationEventListener):
        self._app_listeners.add(listener)

    def publish_event(self, event: ApplicationEvent):
        self._app_event_publisher.publish_event(event)

    def _prepare_bean_factory(self):
        self.add_bean_post_processor(BeanContextAwareProcessor(self))

    def _post_process_bean_factory(self):
        pass

    def _invoke_bean_factory_post_processors(self):
        def invoke_bean_factory_post_processors(post_processors, factory):
            for post_processor in post_processors:
                post_processor.post_process_bean_factory(factory)

        def invoke_bean_definition_registry_post_processors(post_processors, registry):
            for post_processor in post_processors:
                post_processor.post_process_bean_definition_registry(registry)

        for post_processor in self.bean_factory_post_processors:
            if isinstance(post_processor, BeanDefinitionRegistryPostProcessor):
                post_processor.post_process_bean_definition_registry(self)

        post_processor_beans = list(self.get_beans_of_type(BeanFactoryPostProcessor).values())
        regular_post_processors = []
        registry_post_processors = []
        for post_processor in (post_processor_beans + self.bean_factory_post_processors):
            if isinstance(post_processor, BeanDefinitionRegistryPostProcessor):
                registry_post_processors.append(post_processor)
            regular_post_processors.append(post_processor)
        invoke_bean_definition_registry_post_processors(registry_post_processors, self)
        invoke_bean_factory_post_processors(regular_post_processors, self)

    def _register_bean_post_processors(self):
        def register_bean_post_processors(factory, post_processors):
            for post_processor in post_processors:
                factory.add_bean_post_processor(post_processor)

        post_processor_beans = list(self.get_beans_of_type(BeanPostProcessor).values())
        register_bean_post_processors(self, post_processor_beans)

    def _init_application_event_publisher(self):
        self._app_event_publisher = ApplicationEventPublisher(self)
        self.register_singleton(APPLICATION_EVENT_PUBLISHER, self._app_event_publisher)

    def _register_application_listeners(self):
        for listener in self._app_listeners:
            self._app_event_publisher.add_application_listener(listener)
        bean_names = self.get_bean_names_for_type(ApplicationEventListener)
        for bean_name in bean_names:
            self._app_event_publisher.add_application_listener_bean(bean_name)

    def _finish_bean_factory_initialization(self):
        self.pre_instantiate_singletons()

    def _finish_refresh(self):
        self._app_event_publisher.publish_event(ContextRefreshedEvent(self))

    def _destroy_beans(self):
        self.destroy_singletons()


class AnnotationConfigBeanContext(DefaultBeanContext, AnnotationConfigRegistry):
    def __init__(self):
        DefaultBeanContext.__init__(self)
        self._reader = AnnotatedBeanDefinitionReader(self)
        self._scanner = ModuleBeanDefinitionScanner(self)

    def register(self, *annotated_elements):
        self._reader.register(*annotated_elements)

    def scan(self, *base_modules):
        self._scanner.scan(*base_modules)

    def set_bean_name_generator(self, bean_name_generator):
        self._reader.set_bean_name_generator(bean_name_generator)
        self._scanner.set_bean_name_generator(bean_name_generator)
        self.register_singleton(CONFIGURATION_BEAN_NAME_GENERATOR, bean_name_generator)



from guniflask.config import settings
from guniflask.web import blueprint, get_route


@blueprint
class ConfigController:
    def __init__(self):
        pass

    @get_route('/settings/<name>')
    def get_setting(self, name):
        return {name: settings[name]}



from typing import Optional, Collection, Union

from flask import _request_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.user import User
from guniflask.security.user_details import UserDetails


class SecurityContext:
    AUTHENTICATION = '__authentication'
    USER = '__user'
    DEFAULT_ROLE_PREFIX = 'role_'

    @classmethod
    def get_authentication(cls) -> Optional[Authentication]:
        ctx = _request_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, cls.AUTHENTICATION):
                return None
            return getattr(ctx, cls.AUTHENTICATION)

    @classmethod
    def set_authentication(cls, authentication: Authentication):
        ctx = _request_ctx_stack.top
        if ctx is not None:
            setattr(ctx, cls.AUTHENTICATION, authentication)

    @classmethod
    def get_user(cls) -> Optional[User]:
        ctx = _request_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, cls.USER):
                user = None
                auth = cls.get_authentication()
                if auth is not None:
                    if isinstance(auth, UserAuthentication):
                        if isinstance(auth.principal, UserDetails):
                            user = auth.principal
                        else:
                            user = User(username=auth.name, authorities=auth.authorities)
                setattr(ctx, cls.USER, user)
            return getattr(ctx, cls.USER)

    @classmethod
    def has_authority(cls, user: User, authority: str):
        return cls.has_any_authority(user, [authority])

    @classmethod
    def has_any_authority(cls, user: User, authorities: Collection[str]):
        return cls._has_any_authority_name(user, None, authorities)

    @classmethod
    def has_role(cls, user: User, role: str):
        return cls.has_any_role(user, role)

    @classmethod
    def has_any_role(cls, user, roles: Collection[str]):
        return cls._has_any_authority_name(user, cls.DEFAULT_ROLE_PREFIX, roles)

    @classmethod
    def _has_any_authority_name(cls, user: User, prefix: Union[None, str], roles: Collection[str]):
        def role_with_prefix():
            if role is None or prefix is None:
                return role
            if role.startswith(prefix):
                return role
            return prefix + role

        for role in roles:
            r = role_with_prefix()
            if r in user.authorities:
                return True
        return False


class CurrentUser:
    def __init__(self):
        self._user = SecurityContext.get_user()

    def __getattribute__(self, name):
        if name.startswith('_'):
            return getattr(self._user, name)
        return super().__getattribute__(name)

    def __getattr__(self, name):
        return getattr(self.user, name)

    def has_authority(self, authority):
        return SecurityContext.has_authority(self._user, authority)

    def has_any_authority(self, *authorities):
        return SecurityContext.has_any_authority(self._user, authorities)

    def has_role(self, role):
        return SecurityContext.has_role(self._user, role)

    def has_any_role(self, *roles):
        return SecurityContext.has_any_role(self._user, roles)


current_user: Union[LocalProxy, CurrentUser, User] = LocalProxy(CurrentUser)



from abc import ABCMeta, abstractmethod

from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security_config.security_builder import SecurityBuilder


class ProviderManagerBuilder(SecurityBuilder, metaclass=ABCMeta):
    @abstractmethod
    def with_authentication_provider(self, provider: AuthenticationProvider) -> 'ProviderManagerBuilder':
        pass  # pragma: no cover
